@@ -1,29 +1,17 @@
-import cPickle
 import Cookie
-import hmac
 import os
 import random
 import time
 from datetime import datetime, timedelta
-try:
-    from hashlib import md5, sha1
-except ImportError:
-    from md5 import md5
-    # NOTE: We have to use the callable with hashlib (hashlib.sha1),
-    # otherwise hmac only accepts the sha module object itself
-    import sha as sha1
-
-# Check for pycryptopp encryption for AES
-try:
-    from pycryptopp.cipher import aes
-    from beaker.crypto import generateCryptoKeys
-    crypto_ok = True
-except:
-    crypto_ok = False
+    
+from beaker.crypto import hmac as HMAC, hmac_sha1 as SHA1, md5
+from beaker.util import pickle
 
+from beaker import crypto
 from beaker.cache import clsmap
 from beaker.exceptions import BeakerException
-from beaker.util import b64decode, b64encode, Set
+from base64 import b64encode, b64decode
+
 
 __all__ = ['SignedCookie', 'Session']
 
@@ -37,19 +25,41 @@ class SignedCookie(Cookie.BaseCookie):
     
     def value_decode(self, val):
         val = val.strip('"')
-        sig = hmac.new(self.secret, val[40:], sha1).hexdigest()
-        if sig != val[:40]:
+        sig = HMAC.new(self.secret, val[40:], SHA1).hexdigest()
+        
+        # Avoid timing attacks
+        invalid_bits = 0
+        input_sig = val[:40]
+        if len(sig) != len(input_sig):
+            return None, val
+        
+        for a, b in zip(sig, input_sig):
+            invalid_bits += a != b
+        
+        if invalid_bits:
             return None, val
         else:
             return val[40:], val
     
     def value_encode(self, val):
-        sig = hmac.new(self.secret, val, sha1).hexdigest()
+        sig = HMAC.new(self.secret, val, SHA1).hexdigest()
         return str(val), ("%s%s" % (sig, val))
 
 
 class Session(dict):
-    """Session object that uses container package for storage"""
+    """Session object that uses container package for storage.
+    
+    ``key``
+        The name the cookie should be set to.
+    ``timeout``
+        How long session data is considered valid. This is used 
+        regardless of the cookie being present or not to determine
+        whether session data is still valid.
+    ``cookie_domain``
+        Domain to use for the cookie.
+    ``secure``
+        Whether or not the cookie should only be sent over SSL.
+    """
     def __init__(self, request, id=None, invalidate_corrupt=False,
                  use_cookies=True, type=None, data_dir=None,
                  key='beaker.session.id', timeout=None, cookie_expires=True,
@@ -70,15 +80,20 @@ class Session(dict):
         self.request = request
         self.data_dir = data_dir
         self.key = key
+        
         self.timeout = timeout
         self.use_cookies = use_cookies
         self.cookie_expires = cookie_expires
-        self.cookie_domain = cookie_domain
+        
+        # Default cookie domain/path
+        self._domain = cookie_domain
+        self._path = '/'
         self.was_invalidated = False
         self.secret = secret
         self.secure = secure
         self.id = id
-            
+        self.accessed_dict = {}
+        
         if self.use_cookies:
             cookieheader = request.get('cookie', '')
             if secret:
@@ -89,12 +104,13 @@ class Session(dict):
             else:
                 self.cookie = Cookie.SimpleCookie(input=cookieheader)
             
-            if not self.id and self.cookie.has_key(self.key):
+            if not self.id and self.key in self.cookie:
                 self.id = self.cookie[self.key].value
         
         self.is_new = self.id is None
         if self.is_new:
             self._create_id()
+            self['_accessed_time'] = self['_creation_time'] = time.time()
         else:
             try:
                 self.load()
@@ -110,13 +126,14 @@ class Session(dict):
                               getpid())).hexdigest(), 
         ).hexdigest()
         self.is_new = True
+        self.last_accessed = None
         if self.use_cookies:
             self.cookie[self.key] = self.id
-            if self.cookie_domain:
-                self.cookie[self.key]['domain'] = self.cookie_domain
+            if self._domain:
+                self.cookie[self.key]['domain'] = self._domain
             if self.secure:
                 self.cookie[self.key]['secure'] = True
-            self.cookie[self.key]['path'] = '/'
+            self.cookie[self.key]['path'] = self._path
             if self.cookie_expires is not True:
                 if self.cookie_expires is False:
                     expires = datetime.fromtimestamp( 0x7FFFFFFF )
@@ -135,24 +152,34 @@ class Session(dict):
     def created(self):
         return self['_creation_time']
     created = property(created)
-
-    def delete(self):
-        """Deletes the session from the persistent storage, and sends
-        an expired cookie out"""
-        if self.use_cookies:
-            self._delete_cookie()
-        if hasattr(self, 'namespace'):
-            self.namespace.acquire_write_lock()
-            try:
-                self.namespace.remove()
-            finally:
-                self.namespace.release_write_lock()
+    
+    def _set_domain(self, domain):
+        self['_domain'] = domain
+        self.cookie[self.key]['domain'] = domain
+        self.request['cookie_out'] = self.cookie[self.key].output(header='')
+        self.request['set_cookie'] = True
+    
+    def _get_domain(self):
+        return self._domain
+    
+    domain = property(_get_domain, _set_domain)
+    
+    def _set_path(self, path):
+        self['_path'] = path
+        self.cookie[self.key]['path'] = path
+        self.request['cookie_out'] = self.cookie[self.key].output(header='')
+        self.request['set_cookie'] = True
+    
+    def _get_path(self):
+        return self._path
+    
+    path = property(_get_path, _set_path)
 
     def _delete_cookie(self):
         self.request['set_cookie'] = True
         self.cookie[self.key] = self.id
-        if self.cookie_domain:
-            self.cookie[self.key]['domain'] = self.cookie_domain
+        if self._domain:
+            self.cookie[self.key]['domain'] = self._domain
         if self.secure:
             self.cookie[self.key]['secure'] = True
         self.cookie[self.key]['path'] = '/'
@@ -162,81 +189,117 @@ class Session(dict):
         self.request['cookie_out'] = self.cookie[self.key].output(header='')
         self.request['set_cookie'] = True
 
+    def delete(self):
+        """Deletes the session from the persistent storage, and sends
+        an expired cookie out"""
+        if self.use_cookies:
+            self._delete_cookie()
+        self.clear()
+
     def invalidate(self):
         """Invalidates this session, creates a new session id, returns
         to the is_new state"""
-        if hasattr(self, 'namespace'):
-            namespace = self.namespace
-            namespace.acquire_write_lock()
-            try:
-                namespace.remove()
-            finally:
-                namespace.release_write_lock()
-            
+        self.clear()
         self.was_invalidated = True
         self._create_id()
         self.load()
- 
+    
     def load(self):
         "Loads the data from this session from persistent storage"
-        self.namespace = namespace = self.namespace_class(self.id,
+        self.namespace = self.namespace_class(self.id,
             data_dir=self.data_dir, digest_filenames=False,
             **self.namespace_args)
-        
+        now = time.time()
         self.request['set_cookie'] = True
         
-        namespace.acquire_write_lock()
+        self.namespace.acquire_read_lock()
+        timed_out = False
         try:
             self.clear()
-            now = time.time()
-            
-            if not '_creation_time' in namespace:
-                namespace['_creation_time'] = now
-                self.is_new = True
             try:
-                self.accessed = namespace['_accessed_time']
-                namespace['_accessed_time'] = now
-            except KeyError:
-                namespace['_accessed_time'] = self.accessed = now
+                session_data = self.namespace['session']
+
+                # Memcached always returns a key, its None when its not
+                # present
+                if session_data is None:
+                    session_data = {
+                        '_creation_time':now,
+                        '_accessed_time':now
+                    }
+                    self.is_new = True
+            except (KeyError, TypeError):
+                session_data = {
+                    '_creation_time':now,
+                    '_accessed_time':now
+                }
+                self.is_new = True
             
-            if self.timeout is not None and now - self.accessed > self.timeout:
-                self.invalidate()
+            if self.timeout is not None and \
+               now - session_data['_accessed_time'] > self.timeout:
+                timed_out= True
             else:
-                for k in namespace.keys():
-                    self[k] = namespace[k]
-        
+                # Properly set the last_accessed time, which is different
+                # than the *currently* _accessed_time
+                if self.is_new or '_accessed_time' not in session_data:
+                    self.last_accessed = None
+                else:
+                    self.last_accessed = session_data['_accessed_time']
+                
+                # Update the current _accessed_time
+                session_data['_accessed_time'] = now
+                self.update(session_data)
+                self.accessed_dict = session_data.copy()                
         finally:
-            namespace.release_write_lock()
+            self.namespace.release_read_lock()
+        if timed_out:
+            self.invalidate()
     
-    def save(self):
-        "Saves the data for this session to persistent storage"
+    def save(self, accessed_only=False):
+        """Saves the data for this session to persistent storage
+        
+        If accessed_only is True, then only the original data loaded
+        at the beginning of the request will be saved, with the updated
+        last accessed time.
+        
+        """
+        # Look to see if its a new session that was only accessed
+        # Don't save it under that case
+        if accessed_only and self.is_new:
+            return None
+        
         if not hasattr(self, 'namespace'):
             self.namespace = self.namespace_class(
                                     self.id, 
                                     data_dir=self.data_dir,
                                     digest_filenames=False, 
                                     **self.namespace_args)
-
+        
         self.namespace.acquire_write_lock()
         try:
-            if not '_creation_time' in self.namespace:
-                self.is_new = True
+            if accessed_only:
+                data = dict(self.accessed_dict.items())
+            else:
+                data = dict(self.items())
             
-            keys = self.keys()
-            for k in Set(self.namespace.keys()).difference(keys):
-                del self.namespace[k]
-                    
-            for k in keys:
-                self.namespace[k] = self[k]
-            
-            now = time.time()
-            self.namespace['_accessed_time'] = self['_accessed_time'] = now
-            self.namespace['_creation_time'] = self['_creation_time'] = now
+            # Save the data
+            if not data and 'session' in self.namespace:
+                del self.namespace['session']
+            else:
+                self.namespace['session'] = data
         finally:
             self.namespace.release_write_lock()
         if self.is_new:
             self.request['set_cookie'] = True
     
+    def revert(self):
+        """Revert the session to its original state from its first
+        access in the request"""
+        self.clear()
+        self.update(self.accessed_dict)
+    
+    # TODO: I think both these methods should be removed.  They're from
+    # the original mod_python code i was ripping off but they really
+    # have no use here.
     def lock(self):
         """Locks this session against other processes/threads.  This is
         automatic when load/save is called.
@@ -285,19 +348,21 @@ class CookieSession(Session):
     def __init__(self, request, key='beaker.session.id', timeout=None,
                  cookie_expires=True, cookie_domain=None, encrypt_key=None,
                  validate_key=None, secure=False, **kwargs):
-        if not crypto_ok and encrypt_key:
-            raise BeakerException("pycryptopp is not installed, can't use "
+        
+        if not crypto.has_aes and encrypt_key:
+            raise InvalidCryptoBackendError("No AES library is installed, can't generate "
                                   "encrypted cookie-only Session.")
         
         self.request = request
         self.key = key
         self.timeout = timeout
         self.cookie_expires = cookie_expires
-        self.cookie_domain = cookie_domain
         self.encrypt_key = encrypt_key
         self.validate_key = validate_key
         self.request['set_cookie'] = False
         self.secure = secure
+        self._domain = cookie_domain
+        self._path = '/'
         
         try:
             cookieheader = request['cookie']
@@ -326,6 +391,7 @@ class CookieSession(Session):
             if self.timeout is not None and time.time() - \
                self['_accessed_time'] > self.timeout:
                 self.clear()
+            self.accessed_dict = self.copy()
             self._create_cookie()
     
     def created(self):
@@ -335,18 +401,35 @@ class CookieSession(Session):
     def id(self):
         return self['_id']
     id = property(id)
+
+    def _set_domain(self, domain):
+        self['_domain'] = domain
+        self._domain = domain
+        
+    def _get_domain(self, domain):
+        return self._domain
     
+    domain = property(_get_domain, _set_domain)
+    
+    def _set_path(self, path):
+        self['_path'] = path
+        self._path = path
+    
+    def _get_path(self, domain):
+        return self._path
+    
+    path = property(_get_path, _set_path)
+
     def _encrypt_data(self):
         """Serialize, encipher, and base64 the session dict"""
         if self.encrypt_key:
             nonce = b64encode(os.urandom(40))[:8]
-            encrypt_key = generateCryptoKeys(self.encrypt_key,
+            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
                                              self.validate_key + nonce, 1)
-            ctrcipher = aes.AES(encrypt_key)
-            data = cPickle.dumps(self.copy(), protocol=2)
-            return nonce + b64encode(ctrcipher.process(data))
+            data = pickle.dumps(self.copy(), 2)
+            return nonce + b64encode(crypto.aesEncrypt(data, encrypt_key))
         else:
-            data = cPickle.dumps(self.copy(), protocol=2)
+            data = pickle.dumps(self.copy(), 2)
             return b64encode(data)
     
     def _decrypt_data(self):
@@ -354,24 +437,28 @@ class CookieSession(Session):
         dict"""
         if self.encrypt_key:
             nonce = self.cookie[self.key].value[:8]
-            encrypt_key = generateCryptoKeys(self.encrypt_key,
+            encrypt_key = crypto.generateCryptoKeys(self.encrypt_key,
                                              self.validate_key + nonce, 1)
-            ctrcipher = aes.AES(encrypt_key)
             payload = b64decode(self.cookie[self.key].value[8:])
-            data = ctrcipher.process(payload)
-            return cPickle.loads(data)
+            data = crypto.aesDecrypt(payload, encrypt_key)
+            return pickle.loads(data)
         else:
             data = b64decode(self.cookie[self.key].value)
-            return cPickle.loads(data)
+            return pickle.loads(data)
     
     def _make_id(self):
         return md5(md5(
-            "%f%s%f%d" % (time.time(), id({}), random.random(), getpid())
+            "%f%s%f%s" % (time.time(), id({}), random.random(), getpid())
             ).hexdigest()
         ).hexdigest()
     
-    def save(self):
+    def save(self, accessed_only=False):
         """Saves the data for this session to persistent storage"""
+        if accessed_only and self.is_new:
+            return
+        if accessed_only:
+            self.clear()
+            self.update(self.accessed_dict)
         self._create_cookie()
     
     def expire(self):
@@ -386,7 +473,6 @@ class CookieSession(Session):
             self['_id'] = self._make_id()
         self['_accessed_time'] = time.time()
         
-        
         if self.cookie_expires is not True:
             if self.cookie_expires is False:
                 expires = datetime.fromtimestamp( 0x7FFFFFFF )
@@ -408,12 +494,14 @@ class CookieSession(Session):
             raise BeakerException("Cookie value is too long to store")
         
         self.cookie[self.key] = val
-        if self.cookie_domain:
-            self.cookie[self.key]['domain'] = self.cookie_domain
+        if '_domain' in self:
+            self.cookie[self.key]['domain'] = self['_domain']
+        elif self._domain:
+            self.cookie[self.key]['domain'] = self._domain
         if self.secure:
             self.cookie[self.key]['secure'] = True
         
-        self.cookie[self.key]['path'] = '/'
+        self.cookie[self.key]['path'] = self.get('_path', '/')
         
         if expires:
             self.cookie[self.key]['expires'] = \
@@ -422,12 +510,15 @@ class CookieSession(Session):
         self.request['set_cookie'] = True
     
     def delete(self):
+        """Delete the cookie, and clear the session"""
         # Send a delete cookie request
         self._delete_cookie()
         self.clear()
     
-    # Alias invalidate to delete
-    invalidate = delete
+    def invalidate(self):
+        """Clear the contents and start a new session"""
+        self.delete()
+        self['_id'] = self._make_id()
 
 
 class SessionObject(object):
@@ -488,9 +579,40 @@ class SessionObject(object):
         return self._session().has_key(key)
     
     def get_by_id(self, id):
+        """Loads a session given a session ID"""
         params = self.__dict__['_params']
         session = Session({}, use_cookies=False, id=id, **params)
         if session.is_new:
-            session.namespace.remove()
             return None
         return session
+    
+    def save(self):
+        self.__dict__['_dirty'] = True
+    
+    def delete(self):
+        self.__dict__['_dirty'] = True
+        self._session().delete()
+    
+    def persist(self):
+        """Persist the session to the storage
+        
+        If its set to autosave, then the entire session will be saved
+        regardless of if save() has been called. Otherwise, just the
+        accessed time will be updated if save() was not called, or
+        the session will be saved if save() was called.
+        
+        """
+        if self.__dict__['_params'].get('auto'):
+            self._session().save()
+        else:
+            if self.__dict__.get('_dirty'):
+                self._session().save()
+            else:
+                self._session().save(accessed_only=True)
+    
+    def dirty(self):
+        return self.__dict__.get('_dirty', False)
+    
+    def accessed(self):
+        """Returns whether or not the session has been accessed"""
+        return self.__dict__['_sess'] is not None
